from django.contrib.auth.backends import RemoteUserBackend
from django.contrib.auth.models import User
from django_tequila.tequila_client import TequilaClient
from django_tequila.tequila_client.config import EPFLConfig
from django.conf import settings

class TequilaBackend(RemoteUserBackend):
    """
    Authenticate against a Tequila server, using the TequilaClient class
    This backend is to be used in conjunction with the ``RemoteUserMiddleware``
    found in the middleware module of this package, and is used when the server
    is handling authentication outside of Django.

    By default, the ``authenticate`` method creates ``User`` objects for
    usernames that don't already exist in the database.  Subclasses can disable
    this behavior by setting the ``create_unknown_user`` attribute to
    ``False``.
    """
    create_unknown_user = True
    
    """ Created user should be inactive by default ?"""
    try:
        set_created_user_at_inactive = settings.TEQUILA_NEW_USER_INACTIVE
    except AttributeError:
        set_created_user_at_inactive = False
    
    """ Set server url"""
    try:
        tequila_server_url = settings.TEQUILA_SERVER_URL
    except AttributeError:
        tequila_server_url = "" 
    
    def authenticate(self, tequila_key):
        """
        The username passed as ``remote_user`` is considered trusted.  This
        method simply returns the ``User`` object with the given username,
        creating a new ``User`` object if ``create_unknown_user`` is ``True``.

        Returns None if ``create_unknown_user`` is ``False`` and a ``User``
        object with the given username is not found in the database.
        """
        if not tequila_key:
            return
        user = None

        if self.tequila_server_url:
            user_attributes = TequilaClient(EPFLConfig(server_url = self.tequila_server_url)).get_attributes(tequila_key)
        else:
            user_attributes = TequilaClient(EPFLConfig()).get_attributes(tequila_key)
        username = user_attributes['username']

        #keep only the first username, not the user@unit or the multiple users
        if username.find(","):
            username = username.split(",")[0]
            if username.find("@"):
                username = username.split("@")[0]

        # Note that this could be accomplished in one try-except clause, but
        # instead we use get_or_create when creating unknown users since it has
        # built-in safeguards for multiple threads.
        if self.create_unknown_user:
            user, created = User.objects.get_or_create(username=username)
            if created:
                user = self.configure_user(user)
        else:
            try:
                user = User.objects.get(username=username)
            except User.DoesNotExist:
                pass

        if user:
            self.update_attributes_from_tequila(user, user_attributes)

        return user
    
    def update_attributes_from_tequila(self, user, user_attributes):
        """ Fill the user profile with tequila attributes """
        try:
            user.profile.sciper = user_attributes.get('uniqueid')
            user.profile.where = user_attributes.get('where')
            user.profile.unit = user_attributes.get('allunits')
            user.profile.group = user_attributes.get('group')
            user.profile.classe = user_attributes.get('classe')
            user.profile.statut = user_attributes.get('statut')
            user.profile.memberof = user_attributes.get('memberof')
            user.profile.save()
        except Exception:
            pass
        
        ###
        #check for create or update field part
        if user_attributes['firstname']:
            # try a manual truncate if necessary, else allow the truncate warning to be raised
            if len(user_attributes['firstname']) > user._meta.get_field('last_name').max_length and user_attributes['firstname'].find(',') != -1:
                first_name_formatted = user_attributes['firstname'].split(',')[0]
            else:            
                first_name_formatted = user_attributes['firstname']
            
            if user.first_name:
                # need update ?
                if user.first_name != first_name_formatted:
                    user.first_name = first_name_formatted
            else:
                user.first_name = first_name_formatted
        
        if user_attributes['name']:
            if user.last_name:
                if user.last_name != user_attributes['name']:
                    user.last_name = user_attributes['name']
            else:
                user.last_name = user_attributes['name']
        
        if user_attributes['email']:
            if user.email:
                if user.email != user_attributes['email']:
                    user.email = user_attributes['email']
            else:
                user.email = user_attributes['email']
            
        user.save()
        
    def configure_user(self, user):
        """
        Configures a user after creation and returns the updated user.

        By default, returns the user unmodified.
        """
        if self.set_created_user_at_inactive:
            user.is_active = False
            user.save()
        return user


import logging
import json

from django.conf import settings
from django.contrib import auth
from django.contrib.auth.middleware import PersistentRemoteUserMiddleware
from django.core.exceptions import ImproperlyConfigured

from django.http import HttpResponseRedirect
from django.utils.http import urlencode

logger = logging.getLogger('django_tequila.middleware')

def get_query_string(params, new_params=None, remove=None):
    """ Allow to rewrite params from url """
    if new_params is None: new_params = {}
    if remove is None: remove = []
    p = params.copy()
    for r in remove:
        for k in p.keys():
            if k.startswith(r):
                del p[k]
                
    for k, v in new_params.items():
        if v is None:
            if k in p:
                del p[k]
        else:
            p[k] = v
    
    for k, v in p.items():
        if isinstance(v, (list,tuple)):
            p[k] = v[0]
    
    return '?%s' % urlencode(p)

class TequilaMiddleware(PersistentRemoteUserMiddleware):
    """
    Middleware for utilizing tequila web-server-provided authentication.

    If request.user is not authenticated, then this middleware attempts to
    authenticate with the key passed in the ``key`` request.GET.
    If authentication is successful, the user is automatically logged in to
    persist the user in the session.
    """

    # Name of request key to grab the key from
    header = "key"
    
    def process_request(self, request):
        # AuthenticationMiddleware is required so that request.user exists.
        if not hasattr(request, 'user'):
            return 
            raise ImproperlyConfigured(
                "The Django remote user auth middleware requires the"
                " authentication middleware to be installed.  Edit your"
                " MIDDLEWARE setting to insert"
                " 'django.contrib.auth.middleware.AuthenticationMiddleware'"
                " before the TequilaMiddleware class.")
        # If the user is already authenticated, her is already
        # persisted in the session and we don't need to continue.
        if request.user.is_authenticated():
            return

        try:
            # Django diff : key is provided by GET, not META
            tequila_key = request.GET[self.header]
        except KeyError:
            # If specified header doesn't exist then remove any existing
            # authenticated remote-user, or return (leaving request.user set to
            # AnonymousUser by the AuthenticationMiddleware).
            if self.force_logout_if_no_header and request.user.is_authenticated:
                self._remove_invalid_user(request)
            return
        
        # We are seeing this user for the first time in this session, attempt
        # to authenticate the user.
        logger.debug("First time user found, going for authentication "
                     "with the key %s..." % tequila_key)
        user = auth.authenticate(tequila_key = tequila_key)
        logger.debug("User found and logged : %s" % user.__dict__)
        logger.debug("User profile : %s" % user.profile.__unicode__())

        # deny page if not allowed
        if not user:
            logger.debug("User was not able to be authenticated, "
                         "redirect to the not allowed page")
            return HttpResponseRedirect(settings.LOGIN_REDIRECT_IF_NOT_ALLOWED)
        else:
            # User is valid.  Set request.user and persist user in the session
            # by logging the user in.
            request.user = user
            auth.login(request, user)
            logger.debug("User found and logged : %s" % user)

            try:
                clean_url = settings.TEQUILA_CLEAN_URL
                
                if clean_url:
                    #get the url, remove key and redirect to it
                    cleaned_url = request.path
                    
                    #QueryDict to dict
                    params = dict(request.GET.iterlists())
                    
                    cleaned_url += get_query_string(params, remove=self.header)
                    return HttpResponseRedirect(cleaned_url)

            except AttributeError:
                pass    



from django.contrib.auth.decorators import login_required
from django.http import HttpResponse
from django.template import Context, loader
from django.core.urlresolvers import reverse
from django.utils.safestring import mark_safe

def index(request):
    c = Context({'user' : request.user})
    t = loader.get_template('index.html')
    return HttpResponse(t.render(c))

@login_required
def protected_view(request):
    profile = ""

    if request.user.is_authenticated():
        profile = request.user.profile

    return HttpResponse("Successfully seeing a protected view.")
    
def unprotected_view(request):
    login_url = mark_safe('<a href="%s?next=%s">login url</a>' % (reverse('login_view'), request.path))
    logout_url = mark_safe('<a href="%s?next=%s">logout url</a>' % (reverse('logout'), request.path))
    
    c = Context({'user' : request.user,
                 'login_url' : login_url,
                 'logout_url' : logout_url})
    t = loader.get_template('unprotected_view.html')

    return HttpResponse(t.render(c))


from django.contrib.auth.models import User
from django.db import models

class UserProfile(models.Model):
    #required field
    user = models.OneToOneField(User, related_name="profile")
    
    sciper = models.CharField(max_length=100, null=True, blank=True)
    where = models.CharField(max_length=100, null=True, blank=True)
    units = models.CharField(max_length=300, null=True, blank=True)
    group = models.CharField(max_length=150, null=True, blank=True)
    classe = models.CharField(max_length=100, null=True, blank=True)
    statut = models.CharField(max_length=100, null=True, blank=True)
    memberof = models.CharField(max_length=300, null=True, blank=True)
    
    def __unicode__(self):
        return """  Sciper:    %s
                    where:     %s
                    units:     %s
                    group:     %s
                    classe:    %s
                    statut:    %s
                    memberof:  %s
                """ % (self.sciper,
                       self.where,
                       self.units,
                       self.group,
                       self.classe,
                       self.statut,
                       self.memberof
                       )
                
#Trigger for creating a profile on user creation 
def user_post_save(sender, instance, **kwargs):
    profile, new = UserProfile.objects.get_or_create(user=instance)

models.signals.post_save.connect(user_post_save, sender=User)                


try:
    # For python 2
    import urlparse
except ImportError:
    # For python 3
    from urllib import parse as urlparse

from django.conf import settings
from django.contrib.admin.sites import AdminSite
from django.contrib.auth import REDIRECT_FIELD_NAME
from django.http import HttpResponseForbidden, HttpResponseRedirect, QueryDict
from functools import update_wrapper
from django.views.decorators.cache import never_cache
from django.views.decorators.csrf import csrf_protect

class TequilaAdminSite(AdminSite):
    def admin_view(self, view, cacheable=False):
        """
        Verify that the user is logged, and he has the right to access it
        """
        def inner(request, *args, **kwargs):
            if not request.user.is_authenticated():
                return self.login(request)
            if not self.has_permission(request):
                return self.not_allowed_redirect(request)
            return view(request, *args, **kwargs)
        if not cacheable:
            inner = never_cache(inner)
        # We add csrf_protect here so this function can be used as a utility
        # function for any view, without having to repeat 'csrf_protect'.
        if not getattr(view, 'csrf_exempt', False):
            inner = csrf_protect(inner)
        return update_wrapper(inner, view)

    @never_cache
    def login(self, request, extra_context=None):
        """
        Don't display the login form, use the Tequila login process instead
        """
        # user is not logged, so redirect to the real login
        try:
            login_url = settings.LOGIN_URL
        except AttributeError:
            raise ImproperlyConfigured("You have to set a LOGIN_URL in your settings.py")

        next_url = request.get_full_path()
        login_url_parts = list(urlparse.urlparse(login_url))
        querystring = QueryDict(login_url_parts[4], mutable=True)
        querystring[REDIRECT_FIELD_NAME] = next_url
        login_url_parts[4] = querystring.urlencode(safe='/')
        
        return HttpResponseRedirect(urlparse.urlunparse(login_url_parts))
        
    def not_allowed_redirect(self, request):
        try:
            not_allowed_login_url = settings.LOGIN_REDIRECT_IF_NOT_ALLOWED
            return HttpResponseRedirect(not_allowed_login_url)
        except AttributeError:    
            return HttpResponseForbidden("Access forbidden")


import urllib
import sys
import logging
import json
logger = logging.getLogger('django_tequila.client')

from django.utils.encoding import smart_str


try:
    # For Python 2
    from urllib2 import Request
except ImportError:
    # For Python 3
    from urllib.request import Request

try:
    # For Python 2
    url_encoding_method = urllib.urlencode
except AttributeError:
    # For Python 3
    url_encoding_method = urllib.parse.urlencode

try:
    # For Python 2
    from urllib2 import urlopen
except ImportError:
    # For Python 3
    from urllib.request import urlopen


class TequilaClient(object):
    def __init__(self, config):
        self.config = config
        
        self.is_authenticated = False
        self.request_key = None

    def _open_url(self, url, data = None, additional_values = ""):
        """ hide urllib2 access, return header and response """
        url_values = None
        
        if data:
            url_values = url_encoding_method(data)
        if url_values:
            final_url = url + "?" + url_values + additional_values
        else:
            final_url = url

        req = Request(final_url)

        # TODO: try..except HTTPError or URLError
        logger.debug("Opening url %s..." % final_url)
        response = urlopen(req)

        return response.read()
    
    def _get_key(self):
        """ hide urllib2 access """
        logger.debug("Asking for a key...")

        if self.request_key:
            logger.debug("Key already loaded...")
            return self.request_key

        logger.debug("No key found, fetching a new one...")
        
        if self.config.request:
            list_request = '+'.join(self.config.request)
        else:
            raise ValueError("username attribute is mandatory in request")
       
        params = {'urlacces' : smart_str(self.config.redirect_to),
                  'service' : self.config.service,
                  'allows' : self.config.allows
                 }
        
        if self.config.additional_params:
            params.update(self.config.additional_params)
        
        self.request_key = self._open_url(self.config.server_url + "/cgi-bin/tequila/createrequest", params, "&request=" + list_request)[4:-1]

        logger.debug("Key generated : %s " % self.request_key)

        return self.request_key

    key = property(_get_key)
    
    def login_url(self):

        if sys.version_info < (3,):
            key = self.key
        else:
            key = self.key.decode('UTF-8')
        return self.config.server_url + "/cgi-bin/tequila/auth?requestkey=" + key

    def get_attributes(self, key = None):
        """ return a dictionnary of attributes setted by tequila,
            corresponding with the "request" parameter in config
         """
        logger.debug("Asking for attributes...")

        if key:
            params = {'key' : key}
        elif self.request_key:
            params = {'key' : self._get_key()}
        else:
            raise ValueError()
        
        response = self._open_url(self.config.server_url + "/cgi-bin/tequila/fetchattributes", params)
        
        attributes = {}

        # Python 3 need to decode
        if sys.version_info > (3,):
        	response = response.decode('utf-8')

        list_attributes = response.split('\n')

        logger.debug("Attributes returned by %s : \n%s" % (self.config.server_url,
                                                         json.dumps(list_attributes, sort_keys=True, indent=4)))
        
        for attribute in list_attributes[:-1]:
            splitted_attr = attribute.split('=')

            if len(splitted_attr) == 1:
                splitted_attr.append('')

            attributes[splitted_attr[0]] = splitted_attr[1]

        if not self._verify_attributes(attributes):
            raise StandardError()

        logger.debug("Attributes after formatting and returned: \n%s" % json.dumps(attributes,
                                                                                 sort_keys=True, indent=4))

        return attributes
    
    def _verify_attributes(self, attributes):
        needed_attrs = ['org', 'user', 'host', 'key']
        
        for needed_attribute in needed_attrs:
            try:
                if not attributes[needed_attribute]:
                    return False
            except KeyError:
                return False 
        return True


from django.conf.urls import url
from django_tequila.tequila_auth_views import login, logout, not_allowed

urlpatterns = [
    url(r'^login/?$', login, name='login_view'),
    url(r'^logout/$', logout, name='logout'),
    url(r'^not_allowed/$', not_allowed, name='not_allowed'),
]



# -*- coding: utf-8 -*-
# Generated by Django 1.10 on 2016-08-04 12:02
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='UserProfile',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('sciper', models.CharField(blank=True, max_length=100, null=True)),
                ('where', models.CharField(blank=True, max_length=100, null=True)),
                ('units', models.CharField(blank=True, max_length=300, null=True)),
                ('group', models.CharField(blank=True, max_length=150, null=True)),
                ('classe', models.CharField(blank=True, max_length=100, null=True)),
                ('statut', models.CharField(blank=True, max_length=100, null=True)),
                ('memberof', models.CharField(blank=True, max_length=300, null=True)),
                ('user', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='profile', to=settings.AUTH_USER_MODEL)),
            ],
        ),
    ]



class Config(object):
    def __init__(self,
                 server_url,
                 additional_params = None,
                 redirect_to = None,
                 service = None,
                 request = None,
                 language = None,
                 localserver = None,
                 org = None,
                 langage = None,
                 want_right = None,
                 want_role = None,
                 allows = None,
                 strong_authentication = False
                 ):
        
        self.server_url = server_url
        
        if redirect_to:
            if redirect_to.find('http://') == -1 and redirect_to.find('https://') == -1:
                self.redirect_to = 'http://' + redirect_to
            else:
                self.redirect_to = redirect_to
        else:
            self.redirect_to = None
            
        self.additional_params = additional_params

        if strong_authentication:
            strong_authentication_param = {'authstrength' : 3}
            
            if self.additional_params:
                self.additional_params.update(strong_authentication_param)
            else:
                self.additional_params = strong_authentication_param
        
        self.service = service
        self.request = request
        self.langage = langage or "english"
        self.allows = allows


class EPFLConfig(Config):
    """ Specific config currently used @ EPFL
    More info on config options https://tequila.epfl.ch/info """

    def __init__(self, allow_guests = False, *args, **kwargs):
        if not kwargs.get('server_url'):
            kwargs['server_url'] = "https://tequila.epfl.ch"

        super(EPFLConfig, self).__init__(*args, **kwargs)
        
        if not kwargs.get('org'):
            self.org = "EPFL"
            
        if not kwargs.get('request'):
            self.request = ["name",
                            "firstname",
                            "statut",
                            "username",
                            "classe",
                            "uniqueid",
                            "email",
                            "allunits",
                            "where",
                            "group",
                            "memberof"]
            
        if allow_guests:
            self.allow_guests()
    
    def allow_guests(self):
        """ Use preventive config, by default, set guests has not allowed """
        if self.allows:
            self.allows += '|categorie=epfl-guests'
        else:
            self.allows = 'categorie=epfl-guests'



from setuptools import setup

setup(
    name='django-tequila',
    version=__import__('django_tequila').__version__,
    
    author='Julien Delasoie',
    author_email='julien.delasoie at epfl ch',
    keywords='django, tequila, authentication',

    url='http://kis-doc.epfl.ch/django-tequila/',
    license="LGPLv3",
    description='A Tequila authentication backend for django',
    long_description=open('README', 'r').read(),

    packages = ['django_tequila',
                'django_tequila.django_backend',
                'django_tequila.tequila_client',
                'django_tequila.tequila_auth_views'],

    classifiers=[
        'Programming Language :: Python',
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)',
        'Operating System :: OS Independent',
        'Framework :: Django',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: System :: Systems Administration :: Authentication/Directory',
    ]
)


from django.contrib import admin
from django.contrib.auth.admin import UserAdmin
from django.contrib.auth.models import User
from sample_app.models import UserProfile

class UserProfileInline(admin.StackedInline):
    model = UserProfile
    
class CustomUserAdmin(UserAdmin):
    inlines = [UserProfileInline]

admin.site.unregister(User)
admin.site.register(User, CustomUserAdmin)